/* ****************************************************************************
 * Witty Wizard CMS Example of Sub Menu using a Suckerfish type of Menu System
 */
#include <iostream>
#include <fstream>
#include <boost/lexical_cast.hpp>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include <Wt/WApplication>
#include <Wt/WEnvironment>
#include <Wt/WMenu>
#include <Wt/WMenuItem>
#include <Wt/WPopupMenu>
#include <Wt/WStackedWidget>
#include <Wt/WText>
#include <Wt/WBootstrapTheme>
#include <Wt/WTemplate>
#include <Wt/WNavigationBar>
#include <Wt/WGlobal>
//
/* ****************************************************************************
 * Master Menu
 * name: Name of Menu Item
 * path: Internal Path
 * type: "", "submenu", "topsub", "topend", "subsub", "subend"
 * parent: Desending from here
 * grandparent: Desending from here if sub of a sub menu
 * isSelected: True if path is selected, else false
 */
struct MasterMenu
{
    MasterMenu(const std::string& name, std::string path, std::string type, std::string parent, std::string grandparent, bool isSelected) : name_(name), path_(path), type_(type), parent_(parent), grandparent_(grandparent), isSelected_(isSelected) { }
    std::string name_;
    std::string path_;
    std::string type_;
    std::string parent_;
    std::string grandparent_;
    bool isSelected_;
}; // end struct MasterMenu
/* ****************************************************************************
 * DeferredWidget
 * A utility container widget which defers creation of its single
 * child widget until the container is loaded (which is done on-demand by a WMenu).
 * The constructor takes the create function for the widget as a parameter.
 *
 * We use this to defer widget creation until needed.
 */
template <typename Function>
class DeferredWidget : public Wt::WContainerWidget
{
    public:
        DeferredWidget(Function f) : f_(f) { }
    private:
        void load()
        {
            Wt::WContainerWidget::load();
            if (count() == 0) { addWidget(f_()); }
        }
        Function f_;
}; // end class DeferredWidget
/* ****************************************************************************
 * deferCreate
 */
template <typename Function>
DeferredWidget<Function> *deferCreate(Function f)
{
    return new DeferredWidget<Function>(f);
} // end DeferredWidget
/* ****************************************************************************
 * class App
 */
class App : public Wt::WApplication
{
    public:
        App(const Wt::WEnvironment& env);
        // Menu Vector Array
        std::vector<MasterMenu> mastermenus;
        // Add Master Menu
        void AddMasterMenu(const MasterMenu& l) { mastermenus.push_back(l); }
    private:
        Wt::WTemplate* homeTemplate_ = NULL;
        void CreateMenu();
        void ReInit();
        void logInternalPath(const std::string& path);
        Wt::WWidget* MenuMan();
        std::string thePath;
}; // end class App
/* ****************************************************************************
 * App Contructor
 */
App::App(const Wt::WEnvironment& env) : Wt::WApplication(env)
{
    //
    messageResourceBundle().use(appRoot() + "sub-menu", false);

    Wt::WBootstrapTheme *bootstrapTheme = new Wt::WBootstrapTheme();
    bootstrapTheme->setVersion(Wt::WBootstrapTheme::Version3);
    setTheme(bootstrapTheme);

    useStyleSheet(Wt::WApplication::resourcesUrl() + "themes/wittywizard/blue/SfMenuHoriz.css"); // FIXME: Theme

    internalPathChanged().connect(this, &App::logInternalPath);
    std::string myLocale = env.locale().name().c_str();
    if (myLocale == "en-US")
    {
        // In production, I read this in from an xml file and fill a database, and query for locale
        // Menu-0  Menu-1  Menu-2 (myMainMenu)     Menu-3   Menu-4
        //                    SubMenu-1 (myPopupMenu)           SubMenu-1
        //                       Sub-Popup-1 (myPopupMenu)      SubMenu-2
        //                          SubMenu-2 (myMenuItem)
        /*
         * This menu system requires the path to be consistent, every menu item in a popup needs to be the same,
         * the top level will be some name, i.e. menux, each item will start with menux,
         * each sub off of it will start with menux and add to it menux/menuz.
         * The Parent is the top level ul, the grandparent is for sub menus in a sub menu
         */
        // Add Content to sub-menu.xml with path as id
        //                 name                 path              type       parent           grandparent
        AddMasterMenu(MasterMenu("Menu-0"     , "menu0"          , ""        , ""              , "", false)); // myMainMenu
        AddMasterMenu(MasterMenu("Menu-1"     , "menu1"          , ""        , ""              , "", false)); // myMainMenu

        AddMasterMenu(MasterMenu("Menu-2"     , "menu2"          , "topsub"  , "menu2"         , "", false)); // myPopupMenu[l.path_], myMenuItem[l.path_]

        AddMasterMenu(MasterMenu("SubMenu-1" , "menu2/submenu1" , "submenu" , "menu2"         , "", false)); // myMenuItem[l.path_]
        AddMasterMenu(MasterMenu("SubMenu-2" , "menu2/submenu2" , "submenu" , "menu2"         , "", false)); // myMenuItem[l.path_]

        AddMasterMenu(MasterMenu("Sub-Sub-1", "menu2/subsub1"  , "subsub"  , "menu2/submenu2", "menu2", false)); // myPopupMenu[l.path_]
        AddMasterMenu(MasterMenu("SubMenu-3", "menu2/submenu3" , "submenu" , "menu2/submenu2", "menu2", false)); // myMenuItem[l.path_]
        AddMasterMenu(MasterMenu("Sub-Sub-1", "menu2/subsub1"  , "subend"  , "menu2/submenu2", "menu2", false)); // myMainMenu->addItem(myPopupMenu[l.path_]->setSubMenu(myMenuItem[l.path_]));

        AddMasterMenu(MasterMenu("Menu-2"     , "menu2"          , "topend"  , "menu2"         , "", false)); // myMainMenu->addItem(myPopupMenu[l.path_]->setSubMenu(myMenuItem[l.path_]));

        AddMasterMenu(MasterMenu("Menu-3"     , "menu3"          , ""        , ""              , "", false)); // myMainMenu
        /*
         * This is an example of xml, see Witty Wizard project for details on reading this into a database
            <menuman name="Menu-0"    path="menu0"          type=""        parent=""               grandparent=""      language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
            <menuman name="Menu-1"    path="menu1"          type=""        parent=""               grandparent=""      language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
            <menuman name="Menu-2"    path="menu2"          type="topsub"  parent="menu2"          grandparent=""      language="en" content=""     description=""      keywords="About" title="About" rating="general"></menuman>
            <menuman name="SubMenu-1" path="menu2/submenu1" type="submenu" parent="menu2"          grandparent=""      language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
            <menuman name="SubMenu-2" path="menu2/submenu2" type="submenu" parent="menu2"          grandparent=""      language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
            <menuman name="Sub-Sub-1" path="menu2/subsub1"  type="subsub"  parent="menu2/submenu2" grandparent="menu2" language="en" content=""     description=""      keywords="About" title="About" rating="general"></menuman>
            <menuman name="SubMenu-3" path="menu2/submenu3" type="submenu" parent="menu2/submenu2" grandparent="menu2" language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
            <menuman name="Sub-Sub-1" path="menu2/subsub1"  type="subend"  parent="menu2/submenu2" grandparent="menu2" language="en" content=""     description=""      keywords="About" title="About" rating="general"></menuman>
            <menuman name="Menu-2"    path="menu2"          type="topend"  parent="menu2"          grandparent=""      language="en" content=""     description=""      keywords="About" title="About" rating="general"></menuman>
            <menuman name="Menu-3"    path="menu2"          type=""        parent=""               grandparent=""      language="en" content="Text" description="About" keywords="About" title="About" rating="general"></menuman>
         */
    }
    //else if (myLocale == "cn") { ... }
    //
    ReInit();
} // end App
/* ****************************************************************************
 * Re init
 */
void App::ReInit()
{
    // Clear the Screen
    root()->clear();
    CreateMenu();
} // end ReInit
/* ****************************************************************************
 * Create Menu
 */
void App::CreateMenu()
{
    // internalPath always prefixs a /
    thePath = internalPath();
    std::vector<std::string> parts;
    boost::split(parts, thePath, boost::is_any_of("/"));
    // now we can remove the / so it matches the path
    thePath = thePath.substr(thePath.find("/") + 1);
    Wt::log("start") << "CreateMenu : thePath = ->" << thePath << "<-";
    // This tracks menu items
    std::map <std::string, int> menuItemCounter;
    // This is our template
    homeTemplate_ = new Wt::WTemplate(Wt::WString::tr("ww-template"), root()); //  <message id="template">
    // Create a Container
    Wt::WContainerWidget* container = new Wt::WContainerWidget();
    // Create the stack where the contents will be located
    Wt::WStackedWidget* contents = new Wt::WStackedWidget();
    contents->setStyleClass("content");
    // create a top-level menu
    // Setup a Left-aligned menu.
    Wt::WMenu* myMainMenu = new Wt::WMenu(contents, container);
    myMainMenu->setStyleClass("sfhmenu");
    myMainMenu->removeStyleClass("nav"); // top level li of sfhmenu with children
    std::map <std::string, Wt::WSubMenuItem *> myPopupMenu;
    std::map <std::string, Wt::WMenu *> myMenuItem;
    std::string selectedMenuItem = "";
    //
    for (unsigned i = 0; i < parts.size(); i++)
    {
        Wt::log("notice") << "App::CreateHome: parts size = " << parts.size() << " | parts[" << i << "]=" << parts[i];
    }
    // Clear Selected
    for (unsigned i = 0; i < mastermenus.size(); i++)
    {
        mastermenus.at(i).isSelected_ = false;
    }
    // Set Selected Menu Item
    for (unsigned i = 0; i < mastermenus.size(); i++)
    {
        MasterMenu& menu = mastermenus[i];
        if (menu.type_ == "submenu" || menu.type_ == "")
        {
            Wt::log("notice") << "App::CreateHome: submenu: thePath = " << thePath << " ==  menu.path_ = " << menu.path_;
            if (thePath == menu.path_)
            {
                Wt::log("notice") << "App::CreateHome: isSelected_ submenu = " << menu.path_;
                selectedMenuItem = menu.path_;
                mastermenus.at(i).isSelected_ = true;
                break;
            }
        }
    }
    // Set all Popups Selected in path
    if (!selectedMenuItem.empty())
    {
        for (unsigned i = 0; i < mastermenus.size(); i++)
        {
            MasterMenu& menu = mastermenus[i];
            if (menu.type_ == "")
            {
                if (std::find(parts.begin(), parts.end(), menu.path_) != parts.end())
                {
                    Wt::log("notice") << "App::CreateHome: isSelected_ = " << menu.name_;
                    mastermenus.at(i).isSelected_ = true;
                }
            }
            if (menu.type_ == "topsub")
            {
                if (std::find(parts.begin(), parts.end(), menu.path_) != parts.end())
                {
                    Wt::log("notice") << "App::CreateHome: topsub isSelected_ = " << menu.name_;
                    mastermenus.at(i).isSelected_ = true;
                }
            }
            if (menu.type_ == "subsub")
            {
                if (std::find(parts.begin(), parts.end(), menu.path_) != parts.end())
                {
                    Wt::log("notice") << "App::CreateHome: subsub isSelected_ = " << menu.name_;
                    mastermenus.at(i).isSelected_ = true;
                }
            }
            if (menu.type_ == "submenu")
            {
                if (thePath == menu.path_)
                {
                    if (std::find(parts.begin(), parts.end(), menu.path_) != parts.end())
                    {
                        Wt::log("notice") << "App::CreateHome: submenu isSelected_ = " << menu.name_;
                        mastermenus.at(i).isSelected_ = true;
                    }
                    break;
                }
            }
        }
    }
    // Now lets Build the Menu
    for (unsigned i = 0; i < mastermenus.size(); i++)
    {
        MasterMenu& menu = mastermenus[i];
        if (menu.type_ == "")
        {
            // Add to Main Menu
            myMainMenu->addItem(menu.name_, deferCreate(boost::bind(&App::MenuMan, this)))->setLink(Wt::WLink(Wt::WLink::InternalPath, "/" + menu.path_));
            if (menu.isSelected_)
            {
                Wt::log("notice") << "App::CreateHome: isSelected_ myMainMenu->addItem(" << menu.name_;
                myMainMenu->itemAt(menuItemCounter["rootmenu"]++)->setStyleClass("selectedmm"); // top level li of sfhmenu selected
            }
            else
            {
                Wt::log("notice") << "App::CreateHome: myMainMenu->addItem(" << menu.name_;
                myMainMenu->itemAt(menuItemCounter["rootmenu"]++)->setStyleClass("mainmenumm"); // top level li of sfhmenu
            } // end if (menu.isSelected_)
        }
        else
        {
            if (menu.type_ == "topsub")
            {
                if(myPopupMenu.find(menu.path_) == myPopupMenu.end())
                {
                    myPopupMenu[menu.path_] = new Wt::WSubMenuItem(menu.name_); // , "/" + l.path_ , container
                    myPopupMenu[menu.path_]->removeStyleClass("dropdown"); // top level li of sfhmenu with children
                    if (menu.isSelected_)
                    {
                        Wt::log("notice") << "App::CreateHome: isSelected_ topsub myPopupMenu[" << menu.path_ << "] |  new WSubMenuItem(" << menu.name_;
                        myPopupMenu[menu.path_]->setStyleClass("activemmwc"); // top level li of sfhmenu with children selected
                    }
                    else
                    {
                        Wt::log("notice") << "App::CreateHome: topsub myPopupMenu[" << menu.path_ << "] |  new WSubMenuItem(" << menu.name_;
                        myPopupMenu[menu.path_]->setStyleClass("mmwc"); // top level li of sfhmenu with children
                    } // end if (menu.isSelected_)
                    menuItemCounter["rootmenu"]++;
                } // end if(myPopupMenu.find(menu.path_) == myPopupMenu.end())
                if(myMenuItem.find(menu.path_) == myMenuItem.end())
                {
                    Wt::log("notice") << "App::CreateHome: topsub new WMenu myMenuItem[" << menu.path_;
                    myMenuItem[menu.path_] = new Wt::WMenu(contents, container);
                    myMenuItem[menu.path_]->setStyleClass("sfhmenu"); // top level ul under sfhmenu, container
                } // end if(myMenuItem.find(menu.path_) == myMenuItem.end())
            } // end if (l.type_ == "topsub")
            if (menu.type_ == "subsub")
            {
                if(myPopupMenu.find(menu.parent_) == myPopupMenu.end())
                {
                    myPopupMenu[menu.parent_] = new Wt::WSubMenuItem(menu.name_); // , "/" + l.path_ , container
                    if (menu.isSelected_)
                    {
                        Wt::log("notice") << "App::CreateHome: isSelected_ subsub myPopupMenu[" << menu.parent_ << "] |  new WSubMenuItem(" << menu.name_;
                        myPopupMenu[menu.parent_]->setStyleClass("selectedsm"); // sub level li of sfhmenu selected
                    }
                    else
                    {
                        Wt::log("notice") << "App::CreateHome: subsub myPopupMenu[" << menu.parent_ << "] |  new WSubMenuItem(" << menu.name_;
                        myPopupMenu[menu.parent_]->setStyleClass("smwc"); // sub level li of sfhmenu
                    } // end if (menu.isSelected_)
                } // end if(myPopupMenu.find(menu.parent_) == myPopupMenu.end())
                if(myMenuItem.find(menu.path_) == myMenuItem.end())
                {
                    Wt::log("notice") << "App::CreateHome: subsub new WMenu myMenuItem[" << menu.parent_;
                    myMenuItem[menu.parent_] = new Wt::WMenu(contents, container);
                    myMenuItem[menu.parent_]->setStyleClass("sfhmenu"); // sub level ul of sfhmenu
                } // end if(myMenuItem.find(menu.path_) == myMenuItem.end())
            } // end if (l.type_ == "subsub")
            if (menu.type_ == "submenu")
            {
                myMenuItem[menu.parent_]->addItem(menu.name_, deferCreate(boost::bind(&App::MenuMan, this)))->setLink(Wt::WLink(Wt::WLink::InternalPath, "/" + menu.path_));
                if (menu.isSelected_)
                {
                    Wt::log("notice") << "App::CreateHome: isSelected_ submenu myMenuItem[" << menu.parent_ << "]->addItem(" << menu.name_;
                    myMenuItem[menu.parent_]->itemAt(menuItemCounter[menu.parent_]++)->setStyleClass("selectedsm"); // li of sfhmenu
                }
                else
                {
                    Wt::log("notice") << "App::CreateHome: submenu myMenuItem[" << menu.parent_ << "]->addItem(" << menu.name_;
                    myMenuItem[menu.parent_]->itemAt(menuItemCounter[menu.parent_]++)->setStyleClass("submenumm"); // li of sfhmenu
                } // end if (menu.isSelected_)
            } // end if (l.type_ == "submenu")
            if (menu.type_ == "topend")
            {
                Wt::log("notice") << "App::CreateHome: topend addItem myMainMenu->addItem(myPopupMenu[" << menu.parent_ << "]->setSubMenu(myMenuItem[" << menu.parent_ << "]";
                myPopupMenu[menu.parent_]->setSubMenu(myMenuItem[menu.parent_]);
                myMainMenu->addItem(myPopupMenu[menu.parent_]);
            } // end if (l.type_ == "topend")
        } // end if (l.type_ != "")
        if (menu.type_ == "subend")
        {
            Wt::log("notice") << "App::CreateHome: subend myMainMenu->addItem(myMenuItem[" << menu.grandparent_ << "]->addItem(myPopupMenu[" << menu.parent_ << "])";
            myPopupMenu[menu.parent_]->setSubMenu(myMenuItem[menu.parent_]);
            myMenuItem[menu.grandparent_]->addItem(myPopupMenu[menu.parent_]);
            myMainMenu->addItem(myPopupMenu[menu.grandparent_]);
        } // end if (l.type_ == "subend")
    } // end for (unsigned i = 0; i < thisMenu->mastermenus.size(); i++)
    container->addWidget(contents);
    homeTemplate_->bindWidget("menu", myMainMenu);
    homeTemplate_->bindWidget("contents", container);
} // end CreateMenu
/* ****************************************************************************
 * Menu Manager
 * This function looks up the Page based on its Internal Path
 */
Wt::WWidget* App::MenuMan()
{
    std::string thePath = internalPath();
    thePath = thePath.substr(thePath.find("/") + 1);
    Wt::log("start") << "MenuMan : thePath = ->" << thePath << "<-";
    if (thePath.empty())
        { return new Wt::WText(Wt::WString::tr("page.home")); }
    return new Wt::WText(Wt::WString::tr(thePath));
} // end MenuMan
/* ****************************************************************************
 * log Internal Path
 */
void App::logInternalPath(const std::string& path)
{
    // simulate an access log for the interal paths
    log("Path Change") << path;
    ReInit();
} // end logInternalPath
/* ****************************************************************************
 * Create App
 */
Wt::WApplication* CreateApp(const Wt::WEnvironment& env)
{
    return new App(env);
} // end CreateApp
/* ****************************************************************************
 * main
 */
int main(int argc, char* argv[])
{
    return Wt::WRun(argc, argv, CreateApp);
} // end main
// --- End of File ------------------------------------------------------------
